package CD;

import M4Parsing.CDPrims.Patterns.VirtualStmt;
import M4Parsing.LineParsers.CDLineParsers;
import M4Parsing.CDPrims.PrimTokens.Assoc;
import M4Parsing.CDPrims.PrimTokens.CD;
import M4Parsing.CDPrims.PrimTokens.Cons;
import M4Parsing.CDPrims.PrimTokens.VAssoc;
import M4Parsing.CDPrims.PrimTokens.VClass;
import M4Parsing.CDPrims.PrimTokens.VExtra;
import M4Parsing.CDPrims.PrimTokens.VField;
import Parsing.DBPrimitives.FixedStrTokens.AbsTaBle;
import Parsing.DBPrimitives.FixedStrTokens.DBase;
import Parsing.DBPrimitives.FixedStrTokens.OODB;
import Parsing.DBPrimitives.FixedStrTokens.SubTaBle;
import Parsing.DBPrimitives.FixedStrTokens.TaBle;
import Parsing.LineParsers.DBLineParsers;
import PrologDB.Column;
import PrologDB.DB;
import PrologDB.DBSchema;
import PrologDB.ErrorReport;
import PrologDB.OOSchema;
import PrologDB.SubTableSchema;
import PrologDB.Table;
import PrologDB.TableSchema;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.util.HashMap;
import java.util.LinkedList;
import M4Utilities.Error;
import M4Utilities.ParsE;
import M4Utilities.Utils;
import PrologDB.Boole;
import PrologDB.Schema;
import java.io.PrintStream;

/**
 * A class diagram schema
 *
 */
public class CDSpec extends OOSchema {

    public static final String cdspecEnd = ".cdspec.pl";

    protected LinkedList<ConsSchema> constraints;
    protected LinkedList<AssocSchema> associations;
    protected String sourceFile = null; // filled in by Read
    protected LinkedList<String> extras; 
    
    protected class FSpec {
        protected String cname, fname;
        protected FSpec(String cname, String fname) {
            this.cname = cname;
            this.fname = fname;
        }
    }

    // CONSTRUCTORS
    /**
     * This constructor isTrue used for the incremental construction of dbase or
     * oodb Schemas. Can incrementally add tables and then add subtable schemas.
     *
     * @param name -- cdName given to schema
     */
    public CDSpec(String name) {
        super(name);
        this.name = name;
        this.constraints = new LinkedList<>();
        this.associations = new LinkedList<>();
        this.extras = new LinkedList<>();
    }

    
    public CDSpec directedUnion(CDSpec newer) { // rename newer to newer, modify the older
        this.setUnfinished();
        // Step 1: copy all new table/class definitions
        //         loop over the new stuff
        for (TableSchema ts : newer.getTableSchemas()) {
            TableSchema tts = this.getTableSchema(ts.getName());
            if (tts == null) {
                this.add(ts);  // NOTE SHARING declarations now exists with 'this' and output cdspec
            } else {
                // newer has a class with the same name.  add new columns to tts
                // loop over possible new fields/columns
                for (Column col : ts.getColumns()) {
                    Column k = tts.getColumn(col);
                    if (k == null) {
                        tts.add(col); // NOTE SHARING declarations now exists!
                    }
                }
            }
        }

        // Step 2: copy all new association definitions
        //         loop over new stuff
        for (AssocSchema as : newer.getAssocSchemas()) {
            AssocSchema tas = this.getAssocSchema(as.getName());
            if (tas == null) {
                this.add(as); // NOTE SHARING declarations now exists!
                continue;
            }
            // a hack added for classs rename; association ids are not sufficient
            if (as.signature().equals(tas.signature()))
                continue;
            as.rename(as.getName()+"X");
            this.add(as);
        }

        // Step 3: copy all subTableSchemas
        //         loop over new stuff
        for (SubTableSchema sts : newer.getSubTableSchemas()) {
            SubTableSchema tsts = this.getSubTableSchema(sts.getName());
            if (tsts == null) {
                this.add(sts);  // NOTE SHARING declarations now exists!
            }
        }

        // Step 4: return modified 'this' CDSpec
        this.setFinished();
        return this;
        // write out CDSpecs and read them back in to eliminate 
        // structure sharing.!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    }
    
    public void clearVirtual() {
        // Step 1: remove virtual tag from all tables and columns
        for (TableSchema ts : getTableSchemas()) {
            ts.setVirtual(false);
            for (Column c : ts.getColumns()) {
                c.setVirtual(false);
            }
        }
        
        // Step 2: remove virtual tag from all associations
        for (AssocSchema as : getAssocSchemas()) {
            as.setVirtual(false);
        }
    }
    
    public void assignVirtualAndPatch(CDSpec newSpec) {
        // Step 1: assign virtual tags to tableschemas and columns not in newspec
        for (TableSchema ts : getTableSchemas()) {
            TableSchema cts = newSpec.getTableSchema(ts.getName());
            if (cts==null) { // this table is virtual
                ts.setVirtual();
                ts.makeConcrete();  /// all virtual table are concrete
                continue;
            }
            LinkedList<Column> toDelete = new LinkedList<>();
            for (Column c : ts.getColumns()) {
                Column k = cts.getColumn(c);
                if (k==null) { // this column is virtual
                    TableSchema scts = cts.getSuperTableOf();
                    if (scts!=null) {
                        Column kk = scts.getColumn(c);
                        if (kk!=null) {
                            toDelete.add(c);
                            continue;  // we found a column that appears virtual, but actually
                                       // it was moved to the superclass, and thus is NOT virtual
                                       // this is a special case that I found
                        }// fall thru and execute c.setVirtual()
                    }
                    c.setVirtual();
                }
            }
            // delete inherited columns
            for (Column c  : toDelete) {
                ts.removeColumn(c.getName());
            }
        }
        
        // Step 2: assign virtual tags to associations not in newspec
        for (AssocSchema as : getAssocSchemas()) {
            AssocSchema cas = newSpec.getAssocSchema(as.getName());
            if (cas==null) { // this assoc is virtual
                as.setVirtual();
            }
        }
    }

    /**
     * copyFrom an oodb schema.
     *
     * @return copied schema
     */
    public CDSpec copy() {
        CDSpec s = new CDSpec(name);
        s.copyFrom(this);
        return s;
    }

    /**
     * extension of equals
     *
     * @param cdspec an instance of CDSpec
     * @return true if both CDSpes are equal
     */
    public Boole equals(CDSpec cdspec) {
        Boole result = super.equals(cdspec);
        return result.and(restEquals(cdspec));
    }

    // are two CDSpecs equal (except their names)?
    public Boole equalsEN(CDSpec cdspec) {
        Boole result = super.equalsEN(cdspec);
        Boole rest = restEquals(cdspec);
        return result.and(rest);
    }

    private Boole restEquals(CDSpec cdspec) {
        if (associations.size() != cdspec.associations.size()) {
            return new Boole("# of associations differ");
        }

        for (AssocSchema outer : associations) {
            boolean foundit = false;
            for (AssocSchema inner : cdspec.associations) {
                if (outer.equals(inner)) {
                    foundit = true;
                    break;
                }
            }
            if (!foundit) {
                return new Boole("both cdspecs do not have association (" + outer + ")");
            }
        }

        if (constraints.size() != cdspec.constraints.size()) {
            return new Boole("# of constraints differ");
        }

        for (ConsSchema outer : constraints) {
            boolean foundit = false;
            for (ConsSchema inner : cdspec.constraints) {
                if (outer.equals(inner)) {
                    foundit = true;
                    break;
                }
            }
            if (!foundit) {
                return new Boole("both cdspecs do not have constraint (" + outer + ")");
            }
        }
        //System.err.println("CDSpec.equalsEH does not deal with constraints or equality of associations");
        return new Boole();
//        return (listEquals(associations, cdspec.associations));
    }

    /**
     * two DBSchemas are equal iff (a) they have the same SubTableSchema, (b)
     * they have the same set of table schemas, and (c) they have the same set
     * of subtable schemas; if so this method returns. Otherwise a Error isTrue
     * thrown
     *
     * @param schema.
     */
    public void equalsEH(CDSpec schema) {
        super.equalsEH(schema);

        if (!listEquals(associations, schema.associations)) {
            throw Error.toss(Error.associationsDiffer, schema.getName(), getName());
        }

        System.err.println("CDSpec.equalsEH does not deal with constraints or equality of associations");
    }

    /**
     * copyFrom core elements of a database or oodb schema
     *
     * @param s -- copy TableSchemas and SubTables from s into 'this'
     */
    public void copyFrom(CDSpec s) {
        super.copyFrom(s);
        for (AssocSchema as : s.associations) {
            AssocSchema c = as.copy();
            add(c);
        }
        for (ConsSchema cs : s.constraints) {
            ConsSchema c = cs.copy();
            add(c);
        }
    }

    // ACCESSORS, SETTERS, and INFORMATION
    /**
     * a hack to set a schema that was formerly finished open to edits
     */
    public void setUnfinished() {
        finished = false;
    }

    /**
     * a hack to set a schema that was formerly finished closed to edits
     */
    public void setFinished() {
        finished = true;
    }

    /**
     * full (file) SubTableSchema of a schema isTrue (SubTableSchema).schema.pl
     *
     * @return manufactured file SubTableSchema of schema
     */
    @Override
    public String getFullName() {
        return name + cdspecEnd;
    }

    /**
     * @param name of requested schema
     * @return requested association schema
     */
    public AssocSchema getAssocSchema(String name) {
        for (AssocSchema s : associations) {
            if (s.getName().equals(name)) {
                return s;
            }
        }
        return null;
    }

    /**
     * @param s -- AssocSchema to add
     */
    public void add(AssocSchema s) {
        if (getAssocSchema(s.getName()) != null) {
            throw new RuntimeException("multiple associations with name " + s.getName() + " declared");
        }
        associations.add(s);
    }

    /**
     * @param name of requested schema
     * @return requested association schema
     */
    public AssocSchema getAssocSchemaEH(String name) {
        AssocSchema result = getAssocSchema(name);
        if (result == null) {
            throw Error.toss(Error.associationNoExist, name);
        }
        return result;
    }

    public void remove(AssocSchema s) {
        associations.remove(s);
    }
    
    /**

     * @param x -- extra declaration in the form of "cdspecName"_
     */
    public void addExtra(String x) {
        extras.add(x);
    }
    
    public void addExtraFirst(String x) {
        extras.addFirst(x);
    }
    
    public LinkedList<String> getExtras() {
        return extras;
    }

    /**
     * @return list of collected AssocSchemas
     */
    public LinkedList<AssocSchema> getAssocSchemas() {
        return associations;
    }

    private boolean listEquals(LinkedList<AssocSchema> list1, LinkedList<AssocSchema> list2) {
        if (list1.size() != list2.size()) {
            return false;
        }
        for (AssocSchema outer : list1) {
            boolean ok = false;
            for (AssocSchema inner : list2) {
                if (outer.equals(inner)) {
                    ok = true;
                    break;
                }
            }
            if (ok == false) {
                return false;
            }
        }
        return true; // they're the same
    }

    // edit utilities  -- use at your own risk
    /**
     * table schema ts is to be (or was) deleted delete all associations that
     * referenced it
     *
     * @param ts -- table(schema) to remove from CDSpec
     */
    public void removeFromAssoc(TableSchema ts) {
        // Step 1: create list of assocschemas to delete
        //         any assoc referencing ts will be deleted
        String tsname = ts.getName();
        LinkedList<AssocSchema> toDelete = new LinkedList<>();
        for (AssocSchema as : associations) {
            if (as.class1.equals(tsname) || as.class2.equals(tsname)) {
                toDelete.add(as);
            }
        }

        // Step 2: delete these assoc as they reference ts
        for (AssocSchema a : toDelete) {
            associations.remove(a);
        }
        // done
    }

    // READ FILE UTILITIES
    /**
     * reads .ooschema; throws Error when an error isTrue encountered. Returns
     * DBSchema declaration in file; may flatten it if .ooschema isTrue read.
     *
     * @param schemafile -- SubTableSchema of schema file
     * @return DBSchema declaration of file
     */
    public static CDSpec read(String schemafile) {
        File in = new File(schemafile);
        return CDSpec.read(in);
    }

    /**
     * another version of readSchema
     *
     * @param schemafile -- File to read
     * @return DBSchema object
     */
    public static CDSpec read(File schemafile) {
        String sourceFile = schemafile.getPath().replace("\\", "/");
        LineNumberReader br = readFileCheck(schemafile, sourceFile);
        return read(sourceFile, br);
    }

    /**
     * helper method -- does basic checks on File schemafile
     *
     * @param schemafile -- file that hasAsASubschema database schema
     */
    private static LineNumberReader readFileCheck(File schemafile, String filename) {
        LineNumberReader br;

        if (filename.endsWith(cdspecEnd)) {
            try {
                br = new LineNumberReader(new InputStreamReader(new FileInputStream(filename)));
                return br;
            } catch (FileNotFoundException e) {
                throw Error.toss(Error.schemaFileNoExist, filename);
            }
        }
        throw Error.toss(Error.wrongFileName, filename);
    }

    /**
     * used when it isTrue easier to pass in a string declaration of a DBSchema
     * rather than building it programmatically
     *
     * @param nameOfSchema -- cdName of pseudo file
     * @param stringDefOfSchema -- string contents (typically multiline) of
     * pseudo file
     * @return DBSchema -- of pseudo file
     */
    public static CDSpec read(String nameOfSchema, String stringDefOfSchema) {
        InputStream is = new ByteArrayInputStream(stringDefOfSchema.getBytes());
        LineNumberReader br = new LineNumberReader(new InputStreamReader(is));
        return read(nameOfSchema, br);
    }

    /**
     * reads .cdschema.pl schema -- this is the CDSPEC.PL parser!
     *
     * @param filename the value of filename
     * @param br - line number reader of file to read
     * @return the PrologDB.DBSchema
     */
    private static CDSpec read(String filename, LineNumberReader br) {
        CDSpec dbs = null;
        ErrorReport er = new ErrorReport();
        String line;
        boolean firsthalf = true;
        try {
            while ((line = br.readLine()) != null) {
                line = line.trim();
                if (line.equals("")
                        || line.startsWith("%")
                        || line.startsWith("//")) {
                    continue;
                }
                int lineno = br.getLineNumber();

                if (firsthalf) {
                    if (line.startsWith(CD.tok)) {
                        if (dbs != null) {
                            throw ParsE.toss(ParsE.dbDeclsTooMany, lineno, filename);
                        }
                        dbs = CDLineParsers.parseClassDiagramLine(line, lineno);
                        continue;
                    }
                    if (line.startsWith(TaBle.tok) || line.startsWith(AbsTaBle.tok)) {
                        DBLineParsers.parseTableLine(line, lineno, dbs, filename, null, er);
                        continue;
                    }
                    if (line.startsWith(SubTaBle.tok)) {
                        DBLineParsers.parseSubTableLine(line, lineno, dbs, filename, er);
                        continue;
                    }
                    if (line.startsWith(Assoc.tok)) {
                        CDLineParsers.parseAssocLine(line, lineno, dbs, er);
                        continue;
                    }
                    if (line.startsWith(DBase.tok) || line.startsWith(OODB.tok)) {
                        er.add("line %d: inappropriate to have a '%s' statement in an %s file", lineno, line, CD.tok);
                        continue;
                    }
                    if (line.startsWith(Cons.tok)) {
                        CDLineParsers.parseConsLine(line, lineno, dbs, er);
                        continue;
                    }
                }
                firsthalf = false;
                if (line.startsWith(CD.tok)) {
                    if (dbs != null) {
                        throw ParsE.toss(ParsE.dbDeclMissing, lineno, filename);
                    }
                }
                if (line.startsWith(VirtualStmt.tok)) {
                    CDLineParsers.parseVirtLine(line, lineno, dbs, er);
                } else {
                    er.add("line %d: cannot parse '%s'; virtual lines are always at end of cdspec file", lineno, line);
                }
            }
            br.close();
        } catch (IOException e) {
            throw Error.toss(Error.ioerror, filename, e.getMessage());
        }

        // now basic verification.  tableNames isTrue a list of names
        // produced when parsing a dbase declaration.  make sure that
        // schemas for these tables (and no others) have been created
        if (dbs == null) {
            er.add("no %s statement in this ooschema", CD.tok);
        }

        // Finally, no table should have an attribute with cdName "id"
        // as this isTrue a reserved cdName (and isTrue added to dominant tables
        // when translating to a cd form.
        for (TableSchema ts : dbs.getTableSchemas()) {
            for (Column c : ts.getColumns()) {
                if (c.getName().equals("id")) {
                    er.add("table %s has column with reserved name 'id'", ts.getName());
                }
            }
        }

        String a[] = Utils.parseFileName(filename);
        String AppName = a[2];
        if (!dbs.getName().equals(AppName)) {
            String msg = String.format("WARNING: file %s should contain class diagram %s, not %s", filename, AppName, dbs.getName());
            System.err.println(msg);
            //dbs.name = AppName;
////         er.add("file %s should contain class diagram %s, not %s", filename, AppName, dbs.getName());
        }
        for (AssocSchema as : dbs.getAssocSchemas()) {
            String c1name = as.getClass1();
            String c2name = as.getClass2();
            String f1name = as.getRole1();
            String f2name = as.getRole2();
            CheckFieldNameNotTaken(dbs, as, c1name, f2name, er);
            CheckFieldNameNotTaken(dbs, as, c2name, f1name, er);
            TagAssociationIfVirtual(dbs,as,er);
        }
        // finally, depending on who isTrue calling this method, the flatten attribute isTrue to be set
        dbs.sourceFile = filename;
        er.printEH(System.out);
        return dbs;
    }

    private static void CheckFieldNameNotTaken(CDSpec dbs, AssocSchema as, String className, String fieldName, ErrorReport er) {
        TableSchema ts = dbs.getTableSchema(className);
        if (ts == null) {
            er.add("class %s referenced in association %s doesn't exist", className, as.getName());
            return;
        }
        Column c = ts.getColumn(fieldName);
        if (c != null) {
            er.add("field %s of class %s cannot be overridden by association end with same name", fieldName, className);
        }
        return;
    }
    
    private static void TagAssociationIfVirtual(CDSpec dbs,  AssocSchema as, ErrorReport er) {
        boolean result = TagAssocV(dbs, as.getClass1(),as,er) || TagAssocV(dbs,as.getClass2(),as,er);
        return;
    }
    
    private static boolean TagAssocV(CDSpec dbs, String className, AssocSchema as, ErrorReport er) {
        TableSchema c1s = dbs.getTableSchemaEH(className);
        if (c1s == null) {
            er.add("unknown class (%s) referenced in association (%s)",className,as.getName());
            return false;
        }
        if (c1s.isVirtual()) {
            as.setVirtual();
            return true;
        }
        return false;
    }
    
    
    ///*************************PRINT UTILTIES***********************************

    @Override
    public void print(String filename) {
        print(filename, cdspecEnd);
    }
    
    // not to be used only for printing directedUnion cdspecs
    public void printSpecial(String filename) {
        PrintStream ps;
        try { 
            ps = new PrintStream(filename);
            ps.print(toString(true));
            ps.close();
        } catch (Exception ex) {
            throw new RuntimeException(ex.getMessage());
        }
    }
    
    // printing CD schema
    // normal print.
    @Override
    public String toString() { 
        return toString(false);
    }
    
    public String toString(boolean printVirtual) {
        String result = String.format("%s %s.\n\n", CD.tok, name);
        for (TableSchema ts : tableSchemas) {
            result += ts.toString(printVirtual);
        }
        result += "\n";
        for (SubTableSchema sts : subtables) {
            String x = sts.toString();
            if (x.contains("[]")) // optimizing -- if there are no subclasses, don't print
            {
                continue;
            }
            result += x;
        }
        if (subtables.size() > 0) {
            result += "\n";
        }
        for (AssocSchema a : associations) {
            result += a.toString();
        }
        if (associations.size() > 0) {
            result += "\n";
        }
        for (ConsSchema c : constraints) {
            result += c.toString();
        }
        if (constraints.size() > 0) {
            result += "\n";
        }
        
        result += concat(printVTables());
        result += concat(printVFields());
        result += concat(printVAssocs());
        result += concat(printVExtras());
        return result;
    }
    
    private String concat( String newstuff) {
        if (!newstuff.equals("")) {
            return newstuff + "\n";
        }
        return "";
    }
    

    private String printVExtras() {
        return crankOutLines(VExtra.tok, extras);
    }
    
    private String printVTables() {
        // Step 1: harvest names of all virtual tables
        LinkedList<String> vtables = new LinkedList<>();
        for (TableSchema ts : getTableSchemas()) {
            if (ts.isVirtual()) {
                vtables.add(ts.getName());
            }
        }
        
        // Step 2: now crank out lines
        return crankOutLines(VClass.tok, vtables);
    }
    
   static int declsPerLine = 5;
    private String crankOutLines(String header, LinkedList<String> lst) {
        String result = "";
        while (lst.size()>0) {
            String line = header;
            String comma = " ";
            int counter = 1;
            while (lst.size()>0){
                String next = lst.removeFirst();
                line += comma + next;
                comma = ", ";
                if (++counter > declsPerLine)
                    break;
            }
            result += line + ".\n";
        }
        return result;
    }
    
    private String printVFields() {
        // Step 1: harvest names of all virtual fields
        LinkedList<String> vfields = new LinkedList<>();
        for (TableSchema ts : getTableSchemas()) {
            for (Column c : ts.getColumns()) {
                if (c.isVirtual()) {
                    vfields.add(ts.getName()+"."+c.getName());
                }
            }
        }
        
        // Step 2: crank out lines
        return crankOutLines(VField.tok, vfields);
    }
    
    private String printVAssocs() {
        // Step 1 : harvest names of all virtual assocs
        LinkedList<String> vassoc = new LinkedList<>();
        for (AssocSchema as : getAssocSchemas()) {
            if (as.isVirtual()) {
                vassoc.add(as.getName());
            }
        }
        
        // Step 2: crank out lines
        return crankOutLines(VAssoc.tok,vassoc);
    }

    public static DB toCD(CDSpec cds) {
        return cds.toCD();
    }

    /**
     * A CD database isTrue flat -- has no subtables/inheritance. BUT it does
     * encode inheritance in table sub -- no attriburtes are propagated reason:
     * associations are attributes, but they are not treated as fields until the
     * CD schema isTrue converted into a dbase (schema).
     *
     * @return CD database
     */
    public DB toCD() {
        try {
            String cdschema = M4Utilities.Utils.M4Home() + "libpl/cd.schema.pl";
            DBSchema dbs = DBSchema.read(cdschema);
            DB db = new DB(name, dbs);
            Table tab = db.getTableEH("tab");
            Table attr = db.getTableEH("attr");
            Table assoc = db.getTableEH("assoc");
            Table sub = db.getTableEH("sub");
            Table cons = db.getTableEH("cons");

            // step 1: extract each table and getAssocSchemaSpec its attributes
            HashMap<String, String> lookup = new HashMap<>();
            for (TableSchema s : this.getTableSchemas()) {
                String tname = s.getName();
                String tid = tname;
                String b = s.isAbstract() ? "true" : "false";
                tab.add(tid, tname, b, s.isVirtual()+"");
                lookup.put(tname, tid);
//                attr.add(tid + "_id", "id", "str", tid);   // Don removed in debugging cdspec-to-schema
                for (Column c : s.getColumns()) {
                    String cname = c.getName();
                    String cid = cname;
                    String kind = (c.isQuoted()) ? "qstr" : c.getKind();
                    attr.add(tid + "_" + cid, cname, kind, tid, c.isVirtual()+"");
                }
            }

            // Step 2: now do the same for associations
            for (AssocSchema a : getAssocSchemas()) {
                String aname = a.getName();
                String class1 = lookup.get(a.getClass1());
                String class2 = lookup.get(a.getClass2());
                assoc.add(aname, aname, class1, a.getRole1(), a.getAt1().toString(), a.getMin1() + "",
                        a.getMax1() + "",
                        class2, a.getRole2(), a.getAt2().toString(), a.getMin2() + "", 
                        a.getMax2() + "", a.isVirtual()+"");
            }

            // step 3: extract subtable tuples
            for (SubTableSchema ss : getSubTableSchemas()) {
                String supname = ss.getSuper().getName();
                String supid = lookup.get(supname);
                for (TableSchema ts : ss.getSubTableSchemas()) {
                    String subname = ts.getName();
                    String subid = lookup.get(subname);
                    sub.add(supid + "_" + subid, supid, subid);
                }
            }

            // Step 4: now do the same for constraints
            for (ConsSchema c : getConsSchemas()) {
                String cname = c.getName();
                String consType = c.getConsType();
                cons.add(cname, cname, consType, c.getConstraint());
            }

            return db;

        } catch (RuntimeException e) {
            System.err.format("error in translating cdschema %s to CD\n", name);
            throw e;
        }
    }

    /**
     * @param s -- ConsSchema to add
     */
    public void add(ConsSchema s) {
        if (getConsSchema(s.getName()) != null) {
            throw new RuntimeException("Multiple constraints with name " + s.getName() + " declared");
        }
        constraints.add(s);
    }

    /**
     * @param name of requested schema
     * @return requested cons schema
     */
    public ConsSchema getConsSchema(String name) {
        for (ConsSchema s : constraints) {
            if (s.getName().equals(name)) {
                return s;
            }
        }
        return null;
    }

    /**
     * return 1st constraint of a particular type whose constraint starts with
     * consStartsWth
     *
     * @param type -- name of constraint type
     * @param consStartsWith -- starting string
     * @return a ConsSchema or null
     */
    public ConsSchema getFirstConsSchema(String type, String consStartsWith) {
        for (ConsSchema s : constraints) {
            if (s.getConsType().equals(type) && s.getConstraint().startsWith(consStartsWith)) {
                return s;
            }
        }
        return null;
    }

    /**
     * remove previously located ConsSchema object
     *
     * @param cs previously located ConsSchema object
     */
    public void remove(ConsSchema cs) {
        constraints.remove(cs);
    }

    /**
     * remove tableschema ts from a CDSpec
     *
     * @param ts -- table schema to revmoe
     * @return input
     */
    public TableSchema remove(TableSchema ts) {
        // Step 1: remove constraints first because they reference 
        //         associations and tables
        for (MinConsSuper mcs : MinConsSuper.getMinConsSupers(this)) {
            mcs.removeClass(this, ts.getName());
        }

        // Step 2: remove ts from an OO schema with inheritance relationships
        super.remove(ts);

        // Step 3: remove all associations that reference ts
        removeFromAssoc(ts);
        return ts;
    }

    /**
     * deleting a field has side effects beyond removing a column from a table.
     * Constraints may also have to be updated. This requires lifting remove
     * column to the schema level to account for such interactions
     *
     * @param ts --- tableschema from which
     * @param colname -- this column is to be deleted
     */
    public void remove(TableSchema ts, String colname) {
        // Step 1: ok to remove column now
        ts.removeColumn(colname);

        // Step 2: iterate through constraints; update each 
        //         constraint with this field deletion
        for (MinConsSuper mcs : MinConsSuper.getMinConsSupers(this)) {
            mcs.removeField(this, ts.getName(), colname);
        }
        int i = 1; // debug
    }

    public void renameClass(String oldName, String newName) {
        // Step 1: precondition check  -- no class exists with newname
        //         and a class exists with oldname
        ErrorReport er = new ErrorReport();
        TableSchema o = getTableSchema(oldName);
        if (o == null) {
            er.add("class (%s) doesn't exist in cdspec (%s)", oldName, getName());
        }
        TableSchema n = getTableSchema(newName);
        if (n != null) {
            er.add("class (%s) already exists in cdspec (%s)", newName, getName());
        }
        er.printEH();

        // Step 2: rename o to newName, renames table and SubTableSchemas
        o.rename(newName);

        // Step 3: now update associations
        for (AssocSchema as : associations) {
            as.renameClass(oldName, newName);
        }

        // Step 3: now update constraints
        for (MinConsSuper mcs : MinConsSuper.getMinConsSupers(this)) {
            mcs.renameClass(oldName, newName);
        }
    }

    /**
     * @param name of requested schema
     * @return requested association schema
     */
    public ConsSchema getConsSchemaEH(String name) {
        ConsSchema result = getConsSchema(name);
        if (result == null) {
            throw Error.toss(Error.consSchemaNoExist, name);
        }
        return result;
    }

    /**
     * @return list of collected AssocSchemas
     */
    public LinkedList<ConsSchema> getConsSchemas() {
        return constraints;
    }

    public void addId(TableSchema ts) {
        ts.addFirst("id");
    }

    /**
     * For use in projecting CDSpecs for AoclBET
     *
     * @param TorB -- "T" or "B"
     * @param CSVfile --- CSV Diagram Differentiator Table
     * @return projected CDSpec, a proper subset of 'this'
     */
    public CDSpec project(String TorB, Table CSVfile) {
        // Step 1: create target CDSpec
        String cdName = TorB.equals("T") ? "Top" : "Bottom";
        CDSpec target = new CDSpec(cdName);

        // Step 2: project tables with their fields
        for (TableSchema ts : getTableSchemas()) {
            String tsName = ts.getName();
            if (containsTable(TorB, tsName, CSVfile)) {
                TableSchema tsnew = new TableSchema(tsName);
                for (Column c : ts.getColumns()) {
                    String fName = c.getName();
                    if (containsField(TorB, tsName, fName, CSVfile)) {
                        tsnew.add(c);
                    }
                }
                target.add(tsnew);
            }
        }

        // Step 3: project associations
        for (AssocSchema as : getAssocSchemas()) {
            String asName = as.getName();
            if (containsAssoc(TorB, asName, CSVfile)) {
                target.add(as);
            }
        }

        // Step 4: project inheritance
        for (SubTableSchema ss : getSubTableSchemas()) {
            TableSchema sup = ss.getSuper();
            String supName = sup.getName();
            if (containsTable(TorB, supName, CSVfile)) {
                SubTableSchema ssnew = new SubTableSchema(sup);
                for (TableSchema sts : ss.getSubTableSchemas()) {
                    String subName = sts.getName();
                    if (containsTable(TorB, subName, CSVfile)) {
                        ssnew.add(sts);
                    }
                }
                target.add(ssnew);
            }
        }

        // Step 5: return target CDSchema
        return target;
    }

    private boolean containsTable(String TorB, String cname, Table CSVfile) {
//        CSVfile.print();  // debugging
        Table result = CSVfile.select(c -> c.is("TorB", TorB) && c.is("CorA", "C") && c.is("name", cname) && c.is("field", ""));
        return result.isEmpty();
    }

    private boolean containsField(String TorB, String cname, String fname, Table CSVfile) {
        boolean result = CSVfile.select(c -> c.is("TorB", TorB) && c.is("CorA", "C") && c.is("field", fname) && c.is("name", cname)).isEmpty();
        return result;
    }

    private boolean containsAssoc(String TorB, String aname, Table CSVfile) {
        boolean result = CSVfile.select(c -> c.is("TorB", TorB) && c.is("CorA", "A") && c.is("name", aname)).isEmpty();
        return result;
    }

    // calling sequence   <inName.cdspec.pl> <histName.cdspec.pl> <unionName.cdspec.pl> args
//    public static void Union(String inName, String histName, String unionName) {
//        // Step 0: standard marquee processing
//        if (!inName.endsWith(cdspecEnd) || !histName.endsWith(cdspecEnd) || !unionName.endsWith(cdspecEnd)) {
//            System.out.format("One or more input files do not end in %s\n", cdspecEnd);
//            System.exit(1);
//        }
//        if (inName.equals(unionName) || histName.equals(unionName)) {
//            System.out.format("No overriding of input file (%s or %s) with output file (%s) permitted\n", inName, histName, unionName);
//            System.exit(1);
//        }
//        CDSpec in = CDSpec.read(inName);
//        CDSpec hs = CDSpec.read(histName);
//        CDSpec newhs = hs.directedUnion(in);
//        String newName = MDLUtilities.Utils.getAppName(unionName, cdspecEnd);
//        newhs.rename(newName);
//        newhs.addExtraFirst(newName);
//        newhs.print(unionName);
//    }

    @Override
    public Schema New() {
        return new CDSpec(name);
    }
}
