package M4Parsing.LineParsers;

import M4Parsing.CDPrims.Patterns.AssocStmt;
import M4Parsing.CDPrims.Patterns.CDStmt;
import CD.AssocSchema;
import CD.CDSpec;
import M4Parsing.CDPrims.Patterns.ConsStmt;
import CD.ConsSchema;
import M4Parsing.CDPrims.Patterns.VirtualStmt;
import M4Parsing.CDPrims.PrimTokens.VAssoc;
import M4Parsing.CDPrims.PrimTokens.VClass;
import M4Parsing.CDPrims.PrimTokens.VExtra;
import M4Parsing.CDPrims.PrimTokens.VField;
import Parsing.LineParsers.LineToParse;
import PrologDB.Column;
import PrologDB.ErrorReport;
import PrologDB.TableSchema;
import java.util.LinkedList;

/**
 * this is a simple line-parser for spec files Generally this class is internal
 * and should not be called externally
 */
public class CDLineParsers {

    /**
     * a dbase declaration is on a single line. this method parses the given
     * line @ lineno returning the DBSchema given declared SubTableSchema, also
     * returns a list of tableNames (not their schemas). ParseExceptions thrown
     * if unable to parse
     *
     * @param line -- a string
     * @param lineno -- the line number associated with this line and returned
     * @return DBSchema object with SubTableSchema of schema initialized; also
     * returns list of tableNames -- to be used later in parsing
     */
    public static CDSpec parseClassDiagramLine(String line, int lineno) {
        // Step 1: parse the line
        LineToParse l = new LineToParse(line, lineno);
        CDStmt st = new CDStmt(l);
        st.parse();

        // Step 2: unpack parsing (schemaName, list-of-tables)
        LinkedList<String> list = l.parseList;
        String schemaName = list.removeFirst();
        CDSpec dbs = new CDSpec(schemaName);
        return dbs;
        // note: no subtable declarations are parsed.  So the Schema is unfinished
        // and no associations are parsed
    }

    /**
     * a subtable declaration is on a single line. this method parses the given
     * line at lineno and adds the subtable declaration to DBSchema dbs. A
     * ParseException is thrown if something wrong (parse or semantic error) is
     * discovered.
     *
     * @param line to parse
     * @param lineno of line
     * @param dbs to add subtable definition
     * @param er error reporter
     */
    public static void parseAssocLine(String line, int lineno, CDSpec dbs, ErrorReport er) {
        if (dbs == null) {
            er.add("line %d: subtable declaration before db declaration", lineno);
            return;
        }

        // Step 1: parse the line
        LineToParse l = new LineToParse(line, lineno);
        AssocStmt st = new AssocStmt(l);
        try {
            st.parse();
        } catch (RuntimeException e) {
            er.add(e.getMessage());
            return;
        }

        // Step 2: unpack parsing 
        LinkedList<String> list = l.parseList;
        AssocSchema as = new AssocSchema(list);
        dbs.add(as);

        // Step 3: all associations reference tables.  Make sure that these tables exist!
        isLegal(dbs, as, as.getClass1(), er);
        isLegal(dbs, as, as.getClass2(), er);
    }

    // helper method
    public static void isLegal(CDSpec dbs, AssocSchema a, String className, ErrorReport er) {
        TableSchema ts = dbs.getTableSchema(className);
        if (ts == null) {
            er.add("association (%s %s) references undeclared class %s", a.getClass1(), a.getRole1(), className);
        }
    }

    /**
     * a subtable declaration is on a single line. this method parses the given
     * line at lineno and adds the subtable declaration to DBSchema dbs. A
     * ParseException is thrown if something wrong (parse or semantic error) is
     * discovered.
     *
     * @param line to parse
     * @param lineno of line
     * @param dbs to add subtable definition
     * @param er error reporter
     */
    public static void parseConsLine(String line, int lineno, CDSpec dbs, ErrorReport er) {
        if (dbs == null) {
            er.add("line %d: cons(traint) declaration before db declaration", lineno);
            return;
        }

        // Step 1: parse the line
        LineToParse l = new LineToParse(line, lineno);
        ConsStmt st = new ConsStmt(l);
        try {
            st.parse();
        } catch (RuntimeException e) {
            er.add(e.getMessage());
            return;
        }

        // Step 2: unpack parsing 
        LinkedList<String> list = l.parseList;
        ConsSchema cs = new ConsSchema(list);
        dbs.add(cs);
    }

    public static void parseVirtLine(String line, int lineno, CDSpec dbs, ErrorReport er) {
        if (dbs == null) {
            er.add("line %d: virtual declaration before db declaration", lineno);
            return;
        }

        // Step 1: parse the line
        LineToParse l = new LineToParse(line, lineno);
        VirtualStmt st = new VirtualStmt(l);
        try {
            st.parse();
        } catch (RuntimeException e) {
            er.add(e.getMessage());
            return;
        }

        // Step 2: unpack parsing
        String first = l.parseList.removeFirst();
        switch (first) {
            case VField.tok:
                while (!l.parseList.isEmpty()) {
                    String classname = l.parseList.removeFirst();
                    TableSchema ts = dbs.getTableSchema(classname);
                    if (ts == null) {
                        er.add("Unknown class (%s) referenced in %s statement", classname, VField.tok);
                        l.parseList.removeFirst();
                        continue;
                    }
                    String colName = l.parseList.removeFirst();
                    Column c = ts.getColumn(colName);
                    if (c == null) {
                        er.add("Unknown column (%s) referenced in %s statement", colName, VField.tok);
                    } else {
                        c.setVirtual();
                    }
                }
                break;
            case VClass.tok:
                for (String c : l.parseList) {
                    TableSchema ts = dbs.getTableSchemaEH(c);
                    if (ts == null) {
                        er.add("Unknown class (%s) referenced in %s statement", c, VClass.tok);
                        continue;
                    }
                    ts.setVirtual();
                }
                break;
            case VAssoc.tok:
                for (String a : l.parseList) {
                    AssocSchema as = dbs.getAssocSchemaEH(a);
                    if (a == null) {
                        er.add("Unknown assoc (%s) referenced in %s statement", a, VAssoc.tok);
                        continue;
                    }
                    as.setVirtual();
                }
                break;
            case VExtra.tok:
                for (String a : l.parseList){
                    dbs.addExtra(a);
                }
                break;
            default:
                throw new RuntimeException("unknown virtual declaration : " + first);
        }
    }
}
